'
' MediaMonkey Script
'
' NAME: GenreFinder 4.1
'
' AUTHOR: trixmoto (http://trixmoto.net)
' DATE : 03/08/2009
'
' INSTALL: Copy to Scripts\Auto directory and add the following to Scripts.ini 
'          Don't forget to remove comments (') and set the order appropriately
'
' [GenreFinder]
' FileName=Auto\GenreFinder.vbs
' ProcName=GenreFinder
' Order=50
' DisplayName=&Genre Finder
' Description=Finds track information from Last.Fm
' Language=VBScript
' ScriptType=0 
'
' FIXES: Fixed problems with local searches
'        Fixed webservice calls to use better method
'        Fixed ampersand (&) characters not working in query strings
'        Added option to specify minimum match value (again!)
'        Added ability to use + and - in modified lists (by using ++ and --)
' 

Option Explicit

Sub OnStartup
  'copy previous list
  Dim ini : Set ini = SDB.IniFile
  If ini.StringValue("GenreFinder","Lists") = "" Then
    Dim wht : wht = ini.StringValue("GenreFinder","White")
    Dim blk : blk = ini.StringValue("GenreFinder","Black")
    If wht = "" Then
      If blk = "" Then
        ini.StringValue("GenreFinder","Lists") = "40"
      Else
        ini.StringValue("GenreFinder","Lists") = "42"&blk
      End If
    Else
      ini.StringValue("GenreFinder","Lists") = "41"&wht
    End If
    Call ini.DeleteKey("GenreFinder","White")
    Call ini.DeleteKey("GenreFinder","Black")
  End If  
  
  'add toolbar button
  Dim but : Set but = SDB.Objects("GFToolbarButton")
  If but Is Nothing Then
    Set but = SDB.UI.AddMenuItem(SDB.UI.Menu_TbStandard,0,0)
    but.Caption = "Genre Finder"
    but.IconIndex = 11
    but.UseScript = Script.ScriptPath
    but.OnClickFunc = "Toolbar"
    but.Visible = ini.BoolValue("GenreFinder","Toolbar") 
    Set SDB.Objects("GFToolbarButton") = but
    
    'add option sheet
    Call SDB.UI.AddOptionSheet("GenreFinder Settings",Script.ScriptPath,"InitSheet","SaveSheet",-3)    
  End If
  
  'register event
  Call Script.RegisterEvent(SDB,"OnPlay","Event_OnPlay")    
  
  'create cache
  Set SDB.Objects("GenreFinderCache") = CreateObject("Scripting.Dictionary")
End Sub

Sub Toolbar(but)
  Call GenreFinder()
End Sub

Sub Event_OnPlay()
  If Not (SDB.Player.CurrentSong Is Nothing) Then
    If SDB.IniFile.BoolValue("GenreFinder","OnPlay") Then
      Set SDB.Objects("GenreFinderOnPlay") = SDB.Player.CurrentSong
      Call GenreFinder()
    End If
  End If
End Sub

Sub GenreFinder()
  'check not already running
  Dim list : Set list = SDB.Objects("GenreFinderList")
  If Not (list Is Nothing) Then
    Set SDB.Objects("GenreFinderOnPlay") = Nothing
    Exit Sub
  End If
  
  'check defined fields
  Dim ini : Set ini = SDB.IniFile
  If ini.StringValue("GenreFinder","Lists") = "" Then
    Set SDB.Objects("GenreFinderOnPlay") = Nothing
    Call SDB.MessageBox("GenreFinder: There are no fields defined in the options.",mtError,Array(mbOk))
    Exit Sub
  End If  
  Set SDB.Objects("GenreFinderAutoLists") = CreateObject("Scripting.Dictionary")
  
  'get selected tracks
  Dim onp : onp = 0
  If SDB.Objects("GenreFinderOnPlay") Is Nothing Then
    Set list = SDB.SelectedSongList 
    If list.Count = 0 Then 
      Set list = SDB.AllVisibleSongList        
    End If     
  Else
    onp = 1
    Set list = SDB.NewSongList
    Call list.Add(SDB.Objects("GenreFinderOnPlay"))  
    Set SDB.Objects("GenreFinderOnPlay") = Nothing
  End If
  If list.Count = 0 Then
    Call SDB.MessageBox("GenreFinder: There are no selected tracks to process.",mtError,Array(mbOk))
    Exit Sub
  End If   
  
  'get list depending on run mode
  Dim mode : mode = ini.IntValue("GenreFinder","RunMode")
  Dim str : str = ""
  Select Case mode
    Case 0
      str = "tracks"
    Case 1
      str = "albums"
    Case 2
      str = "artists"
    Case Else
      Call SDB.MessageBox("GenreFinder: Sorry, unknown run mode "&mode&".",mtError,Array(mbOk))
      Exit Sub
  End Select
  Set SDB.Objects("GenreFinderList") = list
  
  'reset logfile
  Dim Debug : Debug = ini.IntValue("GenreFinder","Debug")
  If Debug > 0 Then
    Call clear()
    If onp = 1 Then
      Call out("OnPlay...")
    End If    
    Call out("Runmode:"&str)    
  End If  
  
  'set progress bar
  Dim prog : Set prog = SDB.Progress
  prog.Text = "GenreFinder: Initialising script..."
  prog.Value = 0
  prog.MaxValue = list.Count
  Set SDB.Objects("GenreFinderProgress") = prog
  
  'create queue
  Dim que : Set que = CreateObject("Scripting.Dictionary")
  que.Item("beg") = Timer
  que.Item("tot") = 0
  que.Item("max") = list.Count
  que.Item("fld") = ""
  que.Item("run") = str
  que.Item("deb") = Debug
  que.Item("mat") = ini.IntValue("GenreFinder","Match")
  que.Item("fir") = ini.IntValue("GenreFinder","First")
  If ini.IntValue("GenreFinder","Confirm") > 0 Then
    If (ini.IntValue("GenreFinder","Confirm") = 2) And (onp = 1) Then
      que.Item("con") = 0
    Else
      que.Item("con") = 1
      If Debug > 0 Then
        Call out("Suggestions...")
      End If
    End If
  Else
    que.Item("con") = 0
  End If
  que.Item("onp") = onp
  If ini.BoolValue("GenreFinder","Overwrite") Then
    que.Item("ovr") = 1
    If Debug > 0 Then
      Call out("Overwrite blanks only...")
    End If
  Else
    que.Item("ovr") = 0
  End If  
  If ini.BoolValue("GenreFinder","Multiple") Then
    que.Item("mul") = 1
    If Debug > 0 Then
      Call out("Multiple values...")
    End If
  Else
    que.Item("mul") = 0    
  End If  
  If mode = 0 Then
    que.Item("lib") = 0
  Else
    If ini.BoolValue("GenreFinder","Library") Then
      que.Item("lib") = 1
    Else
      que.Item("lib") = 0
    End If     
  End If
  Dim lst : Set lst = Nothing
  str = ini.StringValue("GenreFinder","Playlist")
  If Not (str = "") Then
    Set lst = SDB.NewSongList
  End If
  Set SDB.Objects("GenreFinderPlaylist") = lst
  que.Item("ply") = str  
  que.Item("art") = ""
  que.Item("aid") = ""  
  Set SDB.Objects("GenreFinderQueue") = que
  
  'loop through list
  Dim cac : Set cac = SDB.Objects("GenreFinderCache")
  Dim i : i = 0
  For i = 0 To list.Count-1  
    que.Item("fld") = ini.StringValue("GenreFinder","Lists")
    prog.Value = i+1 
    Dim itm : Set itm = list.Item(i)
    Dim xml : Set xml = Nothing
    Dim dic : Set dic = Nothing
    Dim boo : boo = True
    Dim fld : fld = ""
    Dim art : art = itm.ArtistName
    If InStr(art,";") > 0 Then
      art = Left(art,InStr(art,";")-1)
    End If    
    que.Item("art") = art
    Select Case mode
      Case 0
        If itm.Title = "" Then
          prog.Text = "GenreFinder: Processing track "&prog.Value&"/"&prog.MaxValue
          boo = False
        Else
          prog.Text = "GenreFinder: Processing track "&prog.Value&"/"&prog.MaxValue&" - Track: '"&itm.Title&"'"
          Set xml = SendQuery("track","artist="&fixurl(itm.ArtistName)&"&track="&fixurl(itm.Title),Debug)
        End If
      Case 1
        If itm.AlbumName = "" Then
          prog.Text = "GenreFinder: Processing track "&prog.Value&"/"&prog.MaxValue
          boo = False
        Else
          prog.Text = "GenreFinder: Processing track "&prog.Value&"/"&prog.MaxValue&" - Album: '"&itm.AlbumName&"'"
          Dim alb : alb = itm.Album.ID
          If cac.Exists("Album"&alb) Then
            If Debug > 2 Then
              Call out("Using cached results...")
            End If
            Set dic = cac.Item("Album"&alb)        
            Call UpdateTrack(dic,que,itm)
            boo = False
          Else
            If que.Item("lib") = 1 Then
              fld = que.Item("fld")
              If Debug > 2 Then
                Call out("Checking library...")
              End If          
              Call LocalSearch(fld,que,itm,alb)
              If fld = "" Then
                boo = False
              Else
                que.Item("fld") = fld
                Set xml = SendQuery("album","artist="&fixurl(itm.ArtistName)&"&album="&fixurl(itm.AlbumName),Debug)
              End If
            Else
              Set xml = SendQuery("album","artist="&fixurl(itm.ArtistName)&"&album="&fixurl(itm.AlbumName),Debug)
            End If 
          End If
        End If
      Case 2
        If art = "" Then
          prog.Text = "GenreFinder: Processing track "&prog.Value&"/"&prog.MaxValue
          boo = False
        Else
          prog.Text = "GenreFinder: Processing track "&prog.Value&"/"&prog.MaxValue&" - Artist: '"&art&"'"
          Dim aid : aid = SDB.Tools.ArtistByName(art).ID
          que.Item("aid") = aid            
          If cac.Exists("Artist"&aid) Then
            If Debug > 2 Then
              Call out("Using cached results...")
            End If
            Set dic = cac.Item("Artist"&aid)        
            Call UpdateTrack(dic,que,itm)
            boo = False
          Else
            If que.Item("lib") = 1 Then
              fld = que.Item("fld")
              If Debug > 2 Then
                Call out("Checking library...")
              End If          
              Call LocalSearch(fld,que,itm,aid)      
              If fld = "" Then
                boo = False
              Else
                que.Item("fld") = fld
                Set xml = SendQuery("artist","artist="&fixurl(art),Debug)
              End If
            Else
              Set xml = SendQuery("artist","artist="&fixurl(art),Debug)
            End If 
          End If
        End If        
    End Select
    If boo Then
      Dim qry : qry = Int(que.Item("tot"))
      que.Item("tot") = qry+1
      If prog.Terminate Then
        If Debug > 0 Then
          Call out("(Cancelled)")
        End If
        Call ClearUp("Process cancelled by user.",mtInformation)
      Else
        If xml Is Nothing Then
          If Debug > 0 Then
            Call out("(Timeout)")
          End If
        Else
          Call ProcessTrack(xml,que,itm)
        End If
      End If        
    End If
  Next

  'unknown list
  Set lst = SDB.Objects("GenreFinderPlaylist")
  If Not (lst Is Nothing) And Not (que.Item("ply") = "") Then
    If lst.Count > 0 Then
      Dim ply : Set ply = SDB.PlaylistByTitle(que.Item("ply"))
      If (ply Is Nothing) Or (ply.Title = "") Then
        If Debug > 1 Then
          Call out("Creating playlist: "&que.Item("ply"))
        End If        
        Set ply = SDB.PlaylistByTitle("").CreateChildPlaylist(que.Item("ply"))  
      End If
      If Not (ply Is Nothing) Then
        If Debug > 1 Then
          Call out("Adding "&lst.Count&" track(s) to playlist: "&que.Item("ply"))
        End If
        Call ply.AddTracks(lst)
      End If
    Else
      If Debug > 2 Then
        Call out("No tracks to add to playlist: "&que.Item("ply"))
      End If              
    End If
  End If
  
  'finish off
  prog.Value = prog.MaxValue
  If Int(que.Item("con")) = 1 Then
    Call ClearUp("[CONF]",mtInformation)
  Else
    Call ClearUp("All "&list.Count&" track(s) have been processed.",mtInformation)
  End If
End Sub

Function SendQuery(run,par,deb)
  Dim prog : Set prog = SDB.Objects("GenreFinderProgress")
  Dim xml : Set xml = CreateObject("Microsoft.XMLHTTP")
  If run = "album" Then
    run = "album.getInfo"
  Else
    run = run&".getTopTags"
  End If
  Dim url : url = "http://ws.audioscrobbler.com/2.0/?method="&run&"&api_key=6cfe51c9bf7e77d6449e63ac0db2ac24&"&par
  If deb > 2 Then
    Call out("@"&url)
  End If  
  Call xml.open("GET",url,true)
  Call xml.send()  
  Dim cnt : cnt = 0
  While (xml.readyState < 4) And (cnt < 300)
    Call SDB.Tools.Sleep(100)
    SDB.ProcessMessages
    cnt = cnt+1
    If prog.Terminate Then
      cnt = 300
    End If    
  WEnd
  If xml.readyState < 4 Then
    Set SendQuery = Nothing
  Else
    Dim str : str = xml.responseText
    If InStr(str,"<lfm status=""ok"">") > 0 Then
      Set xml = CreateObject("Microsoft.XMLDOM")
      xml.LoadXML(str)
      Set SendQuery = xml
    Else
      Set SendQuery = Nothing
    End If
  End If
End Function

Sub ProcessTrack(xml,que,itm)  
  If xml Is Nothing Then
    Exit Sub
  End If
  If que Is Nothing Then
    Exit Sub
  End If
  If itm Is Nothing Then
    Exit Sub
  End If  
  
  'read responses
  Dim Debug : Debug = Int(que.Item("deb"))  
  Dim cac : Set cac = SDB.Objects("GenreFinderCache")
  Dim dic : Set dic = Nothing
  Dim alb : alb = itm.Album.ID
  If cac.Exists("Album"&alb) Then
    Set dic = cac.Item("Album"&alb)
    Call cac.Remove("Album"&alb)
  Else
    Set dic = CreateObject("Scripting.Dictionary")
  End If
  Dim run : run = que.Item("run")
  Dim art : art = UCase(que.Item("art"))
  Dim aid : aid = que.Item("aid")
  Dim fir : fir = Int(que.Item("fir"))
  Dim min : min = Int(que.Item("mat"))
  Dim j : j = 1
  Dim ele,str,fld,res,arr,i,cnt
  For Each ele In xml.getElementsByTagName("tag")
    str = FixGenre(ele.ChildNodes.Item(0).Text)
    cnt = FixCount(ele.ChildNodes.Item(1).Text)
    If (str <> "") And (art <> "") And (InStr(UCase(str),art) > 0) Then
      If Debug > 1 Then
        Call out("'"&str&"' contains artist name ("&que.Item("art")&")")
      End If
      str = ""
    End If
    If (str <> "") And (cnt >= min) Then
      arr = Split(que.Item("fld"),"|")
      For i = 0 To UBound(arr)
        fld = arr(i)  
        res = Acceptable(fld,str)
        If Not (res = "") Then
          fld = FieldToProp(Left(fld,1))
          If dic.Exists(fld) Then
            Dim tmp : tmp = dic.Item(fld)
            If Not IsDup(tmp,res) Then
              dic.Item(fld) = tmp&"; "&res
            End If                
          Else
            Call dic.Add(fld,res)
          End If
        End If
      Next
    Else
      If (Debug > 2) And (str <> "") And (cnt < min) Then
        Call out("'"&str&"' match value ("&cnt&") is below minimum ("&min&")")
      End If      
    End If
    j = j + 1
    If (fir > 0) And (j > fir) Then
      Exit For
    End If
  Next
  
  'cache result
  Select Case run
    Case "albums"
      If Not (alb = "") Then
        Call cac.Add("Album"&alb,dic)
      End If
    Case "artists"
      If Not (aid = "") Then
        Call cac.Add("Artist"&aid,dic)
      End If
  End Select 
  
  'update track
  Call UpdateTrack(dic,que,itm)
End Sub

Sub UpdateTrack(dic,que,itm)  
  'update values    
  Dim i,fld,res,val,pos
  Dim Debug : Debug = Int(que.Item("deb"))  
  Dim upd : upd = False
  Dim con : con = Int(que.Item("con"))
  Dim ovr : ovr = Int(que.Item("ovr"))
  Dim mul : mul = Int(que.Item("mul"))  
  Dim arr : arr = Split(que.Item("fld"),"|")
  For i = 0 To UBound(arr)
    fld = FieldToProp(Left(arr(i),1))
    res = dic.Item(fld)
    If (mul = 0) And (InStr(res,";") > 0) Then
      res = Left(res,InStr(res,";")-1)
      dic.Item(fld) = res
    End If
    If Not (res = "") Then
      Execute("val = itm."&fld)
      If (ovr = 1) And Not (val = "") Then
        If Debug > 1 Then
          Call out(itm.ArtistName&" - "&itm.Title&": "&fld&" not blank ("+val+")")
        End If
      Else
        If (val = res) Then
          If Debug > 1 Then
            Call out(itm.ArtistName&" - "&itm.Title&": "&fld&" is correct ("+res+")")
          End If
        Else
          If con = 0 Then
            Execute("itm."&fld&" = res")
            upd = True
          End If
          If Debug > 0 Then
            Call out(itm.ArtistName&" - "&itm.Title&": "&fld&"="&res)
          End If        
        End If
      End If        
      res = "|"&que.Item("fld")&"|"
      res = Replace(res,"|"&arr(i)&"|","|")
      If res = "|" Then
        que.Item("fld") = ""
      Else
        que.Item("fld") = Mid(res,2,Len(res)-2)
      End If        
    End If
  Next
  
  'store changes
  If upd Then
    Dim list : Set list = SDB.NewSongList
    Call list.Add(itm)
    SDB.Database.BeginTransaction
    Call list.UpdateAll()
    SDB.Database.Commit          
  End If
  
  'add to unknown list
  If Not (que.Item("fld") = "") Then
    Dim lst : Set lst = SDB.Objects("GenreFinderPlaylist")
    If Not (lst Is Nothing) Then
      Call lst.Add(itm)
    End If
  End If
  
  'rest are unknown
  If Debug > 1 Then
    While Not(que.Item("fld") = "")
      fld = que.Item("fld")
      pos = InStr(que.Item("fld"),"|")
      If pos > 0 Then
        fld = Left(que.Item("fld"),pos-1)
        que.Item("fld") = Mid(que.Item("fld"),pos+1)
      Else
        que.Item("fld") = ""
      End If
      fld = FieldToProp(Left(fld,1))
      Call out(itm.ArtistName&" - "&itm.Title&": "&fld&"=["&SDB.Localize("Unknown")&"]")
    WEnd
  End If

  'save for confirmation  
  If con = 1 Then
    Call que.Add("@"&itm.ID,dic)
  End If    
End Sub

Sub ClearUp(mes,typ) 
  Dim Debug : Debug = SDB.IniFile.IntValue("GenreFinder","Debug")
  If Debug > 2 Then
    Call out("Clearing up...")
  End If 
  Dim que : Set que = SDB.Objects("GenreFinderQueue")
  If Not (que Is Nothing) Then
    If Int(que.Item("onp")) = 1 Then
      If Not (mes = "[CONF]") Then
        mes = ""
      End If
    End If
  End If
  If (typ = mtInformation) And (Debug > 0) Then
    If Not (que Is Nothing) Then
      Dim sec : sec = Timer - que.Item("beg")
      If Debug > 1 Then
        Call out("Seconds:"&FormatNumber(sec,2))
      End If
      Dim qry : qry = Int(que.Item("tot"))
      If Debug > 1 Then
        Call out("Queries:"&qry)
      End If
      Dim spq : spq = 0
      If qry > 0 Then
        spq = sec / qry
      End If
      Call out("Sec/Query:"&FormatNumber(spq,2))
    End If
  End If   
  If mes = "[CONF]" Then
    Call Confirmation()
  Else
    Set SDB.Objects("GenreFinderQueue") = Nothing
    Set SDB.Objects("GenreFinderProgress") = Nothing
    Set SDB.Objects("GenreFinderList") = Nothing
    Set SDB.Objects("GenreFinderItem") = Nothing
    Set SDB.Objects("GenreFinderXML") = Nothing
    Set SDB.Objects("GenreFinderPlaylist") = Nothing
    Set SDB.Objects("GenreFinderAutoLists") = Nothing
    If Not (mes = "") Then
      Call SDB.MessageBox("GenreFinder: "&mes,typ,Array(mbOk))
    End If
  End If
End Sub

Function fixurl(sRawURL)
  Const sValidChars = "1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz\/:"
  Dim url : url = Replace(sRawURL,"+","%2B")
  If UCase(Right(url,6)) = " (THE)" Then
    url = "The "&Left(url,Len(url)-6)
  End If
  If UCase(Right(url,5)) = ", THE" Then
    url = "The "&Left(url,Len(url)-5)
  End If  
  
  If Len(url) > 0 Then
    Dim i : i = 1
    Do While i < Len(url)+1
      Dim s : s = Mid(url,i,1)
      If InStr(1,sValidChars,s,0) = 0 Then
        Dim d : d = AscW(s)
        If d = 32 Or d > 65536 Then
          s = "+"
        Else
          If d < 128 Then
            s = DecToHex(d)
          ElseIf d < 2048 Then
            s = DecToUtf(d)
          Else 
            s = DecToUtf2(d)
          End If
        End If
      Else
        Select Case s
          Case "&"
            s = "%2526"
          Case "/"
            s = "%252F"
          Case "\"
            s = "%5C"
          Case ":"
            s = "%3A"
        End Select
      End If
      fixurl = fixurl&s
      i = i + 1
    Loop
  End If
End Function

Function HexToDec(h)
  HexToDec = 0
  Dim i : i = 0
  For i = Len(h) To 1 Step -1
    Dim d : d = Mid(h,i,1)
    d = Instr("0123456789ABCDEF",UCase(d))-1
    If d >= 0 Then
      HexToDec = HexToDec+(d*(16^(Len(h)-i)))
    Else
      HexToDec = 0
      Exit For
    End If
  Next
End Function

Function DecToBin(intDec)
  DecToBin = ""
  Dim d : d = intDec
  Dim e : e = 1024
  While e >= 1
    If d >= e Then
      d = d - e
      DecToBin = DecToBin&"1"
    Else
      DecToBin = DecToBin&"0"
    End If
    e = e / 2
  Wend
End Function

Function DecToBin2(intDec)
  DecToBin2 = ""
  Dim d : d = intDec
  Dim e : e = 65536
  While e >= 1
    If d >= e Then
      d = d - e
      DecToBin2 = DecToBin2&"1"
    Else
      DecToBin2 = DecToBin2&"0"
    End If
    e = e / 2
  Wend
End Function

Function BinToHex(strBin)
  Dim d : d = 0
  Dim i : i = 0
  For i = Len(strBin) To 1 Step -1
    Select Case Mid(strBin,i,1)
      Case "0"
        'do nothing
      Case "1"
        d = d + (2^(Len(strBin)-i))
      Case Else
        BinToHex = "00"
        Exit Function
    End Select
  Next
  BinToHex = DecToHex(d)
End Function

Function DecToHex(d)
  If d < 16 Then
    DecToHex = "%0"&CStr(Hex(d))
  Else
    DecToHex = "%"&CStr(Hex(d))
  End If
End Function

Function DecToUtf(d)
  Dim b : b = DecToBin(d)
  Dim a : a = "110"&Left(b,5)
  b = "10"&Mid(b,6)
  DecToUtf = BinToHex(a)&BinToHex(b)
End Function 

Function DecToUtf2(d)
  Dim b : b = DecToBin2(d)
  Dim a : a = "1110"&Left(b,4)
  Dim c : c = "10"&Mid(b,11,6)
  b = "10"&Mid(b,5,6)
  DecToUtf2 = BinToHex(a)&BinToHex(b)&BinToHex(c)
End Function 

Function FixCount(count)
  If IsNumeric(count) Then
    FixCount = Int(count)
    If FixCount > 100 Then
      FixCount = 100
    End If
  Else
    FixCount = 100
  End If
End Function

Function FixGenre(genre)
  If genre = "" Then
    FixGenre = ""
    Exit Function
  End If
  
  Dim i : i = 0
  Dim arr : arr = Split(genre," ")
  For i = 0 To UBound(arr)
    FixGenre = FixGenre&" "&UCase(Left(arr(i),1))&Mid(arr(i),2)
  Next
  FixGenre = Mid(FixGenre,2)  
End Function

Sub clear()
  Dim wsh : Set wsh = CreateObject("WScript.Shell")
  Dim fso : Set fso = CreateObject("Scripting.FileSystemObject")
  Dim loc : loc = wsh.ExpandEnvironmentStrings("%TEMP%")&"\GenreFinder.log"
  Dim logf : Set logf = fso.CreateTextFile(loc,True)
  logf.Close
End Sub

Sub out(txt)
  Dim wsh : Set wsh = CreateObject("WScript.Shell")
  Dim fso : Set fso = CreateObject("Scripting.FileSystemObject")
  Dim loc : loc = wsh.ExpandEnvironmentStrings("%TEMP%")&"\GenreFinder.log"
  Dim logf : Set logf = fso.OpenTextFile(loc,8,True)
  If SDB.IniFile.IntValue("GenreFinder","Debug") > 3 Then
    logf.WriteLine(Time&" "&SDB.ToAscii(txt))
  Else
    logf.WriteLine(SDB.ToAscii(txt))
  End If
  logf.Close
End Sub

Function IsInList(lst,str)
  IsInList = "[FALSE]"
  If str = "" Or lst = "" Then
    Exit Function
  End If
  
  Dim i,j,t,r,s
  Dim arr : arr = Split(lst,",")
  For i = 0 To UBound(arr)
    t = arr(i)
    j = InStr(t,"=")
    If j = 0 Then
      r = "[TRUE]"
      t = UCase(t)
    Else
      r = Mid(t,j+1)    
      t = UCase(Left(t,j-1))
    End If    
    s = UCase(str)
    If Left(t,1) = "%" Then
      t = Mid(t,2)
      j = Len(t)
      If Len(s) > j Then
        s = Mid(s,Len(s)-j+1)
      End If
    Else
      If Right(t,1) = "%" Then
        j = Len(t)-1
        t = Left(t,j)
        If Len(s) > j Then
          s = Left(s,j)
        End If
      End If
    End If
    If t = s Then
      IsInList = r
      Exit Function
    End If
  Next
End Function  

Sub InitSheet(Sheet)
  Dim ini : Set ini = SDB.IniFile  
  Dim ui : Set ui = SDB.UI
  Dim edt : Set edt = ui.NewCheckBox(Sheet)
  edt.Common.SetRect 5, 10, 200, 20
  edt.Common.ControlName = "GFToolbar"
  edt.Caption = "Show toolbar button?"
  edt.Checked = ini.BoolValue("GenreFinder","Toolbar")
  
  Set edt = ui.NewLabel(Sheet)
  edt.Common.SetRect 5, 38, 50, 20
  edt.Caption = "Show confirmation screen:"
  edt.Autosize = False
  
  Set edt = ui.NewDropDown(Sheet)
  edt.Common.SetRect 140, 35, 75, 20
  edt.Common.ControlName = "GFConfirm" 
  edt.Style = 2
  edt.AddItem("Never")
  edt.AddItem("Always")
  edt.AddItem("Manual")
  edt.ItemIndex = ini.IntValue("GenreFinder","Confirm")    
  
  Set edt = ui.NewCheckBox(Sheet)
  edt.Common.SetRect 5, 60, 200, 20
  edt.Common.ControlName = "GFOnPlay"
  edt.Caption = "Run for current song on play?"
  edt.Checked = ini.BoolValue("GenreFinder","OnPlay") 
  
  Set edt = ui.NewCheckBox(Sheet)
  edt.Common.SetRect 5, 85, 200, 20
  edt.Common.ControlName = "GFOverwrite"
  edt.Caption = "Only overwrite blank values?"
  edt.Checked = ini.BoolValue("GenreFinder","Overwrite") 
  
  Set edt = ui.NewCheckBox(Sheet)
  edt.Common.SetRect 5, 110, 200, 20
  edt.Common.ControlName = "GFMultiple"
  edt.Caption = "Tag with multiple values?"
  edt.Checked = ini.BoolValue("GenreFinder","Multiple")  
  
  Set edt = ui.NewLabel(Sheet)
  edt.Common.SetRect 250, 10, 50, 20
  edt.Caption = "Log level:"
  edt.Autosize = False
  
  Set edt = ui.NewDropDown(Sheet)
  edt.Common.SetRect 335, 7, 100, 20
  edt.Common.ControlName = "GFDebug" 
  edt.Style = 2
  edt.AddItem("0 - None")
  edt.AddItem("1 - Basic")
  edt.AddItem("2 - Extra")
  edt.AddItem("3 - Advanced")
  edt.AddItem("4 - Timings")
  edt.ItemIndex = ini.IntValue("GenreFinder","Debug")  
  
  Set edt = ui.NewLabel(Sheet)
  edt.Common.SetRect 250, 35, 50, 20
  edt.Caption = "Run mode:"
  edt.Autosize = False

  Dim i : i = ini.IntValue("GenreFinder","RunMode")   
  Set edt = ui.NewDropDown(Sheet)
  edt.Common.SetRect 335, 32, 100, 20
  edt.Common.ControlName = "GFRunMode"
  edt.Style = 2
  edt.AddItem("Track")
  edt.AddItem("Album")
  edt.AddItem("Artist")
  edt.ItemIndex = i
  edt.UseScript = Script.ScriptPath
  edt.OnSelectFunc = "Event_OnSelect"  
  
  Set edt = ui.NewLabel(Sheet)
  edt.Common.SetRect 250, 60, 50, 20
  edt.Caption = "Unknown playlist:"
  edt.Autosize = False
  
  Set edt = ui.NewEdit(Sheet)
  edt.Common.SetRect 335, 57, 100, 20
  edt.Common.ControlName = "GFPlaylist"
  edt.Text = ini.StringValue("GenreFinder","Playlist")  
  
  Set edt = ui.NewCheckBox(Sheet)
  edt.Common.SetRect 250, 85, 200, 20
  edt.Common.ControlName = "GFLibrary"
  edt.Caption = "Check library before querying?"
  edt.Checked = ini.BoolValue("GenreFinder","Library")    
  
  Set edt = ui.NewLabel(Sheet)
  edt.Common.SetRect 250, 110, 50, 20
  edt.Caption = "Use top results:"
  edt.Autosize = False
  
  Set edt = ui.NewSpinEdit(Sheet)
  edt.Common.SetRect 335, 107, 100, 20
  edt.Common.ControlName = "GFFirst"
  edt.Common.Hint = "To use all results set to zero"
  edt.MinValue = 0
  edt.MaxValue = 100
  edt.Value = ini.IntValue("GenreFinder","First")
  
  Set edt = ui.NewLabel(Sheet)
  edt.Common.SetRect 250, 135, 50, 20
  edt.Caption = "Minimum match:"
  edt.Autosize = False
  
  Set edt = ui.NewSpinEdit(Sheet)
  edt.Common.SetRect 335, 132, 100, 20
  edt.Common.ControlName = "GFMatch"
  edt.MinValue = 0
  edt.MaxValue = 100
  edt.Value = ini.IntValue("GenreFinder","Match")  
  
  Set edt = ui.NewLabel(Sheet)
  edt.Common.SetRect 5, 135, 100, 20
  edt.Caption = "UPDATE LISTS..."
  edt.Autosize = False
  
  Dim flds : flds = ini.StringValue("GenreFinder","Lists")
  Dim max : max = 10
  i = 0
  While (Len(flds) > 0) And (i < max)
    Dim fld : fld = flds
    Dim pos : pos = InStr(flds,"|")
    If pos > 0 Then
      fld = Left(flds,pos-1)
      flds = Mid(flds,pos+1)
    Else
      flds = ""
    End If 
    i = i + 1
    Set edt = ui.NewDropDown(Sheet)
    edt.Common.SetRect 5, (i*25)+132, 80, 20
    edt.Common.ControlName = "GFProp"&i
    edt.Style = 2
    edt.AddItem(" ")
    edt.AddItem("Custom1")
    edt.AddItem("Custom2")
    edt.AddItem("Custom3")
    edt.AddItem("Genre")
    edt.AddItem("Mood")
    edt.AddItem("Occasion")
    edt.AddItem("Quality")
    edt.AddItem("Tempo")
    edt.AddItem("Custom4")
    edt.AddItem("Custom5")    
    edt.AddItem("Comment")
    edt.ItemIndex = DigIt(Mid(fld,1,1))
    Set edt = ui.NewDropDown(Sheet)
    edt.Common.SetRect 90, (i*25)+132, 60, 20
    edt.Common.ControlName = "GFType"&i
    edt.Style = 2
    edt.AddItem("None")
    edt.AddItem("White")
    edt.AddItem("Black")
    edt.ItemIndex = Int(Mid(fld,2,1))
    Set edt = ui.NewEdit(Sheet)
    edt.Common.SetRect 155, (i*25)+132, 220, 20
    edt.Common.ControlName = "GFText"&i
    If Len(fld) > 2 Then
      edt.Text = Mid(fld,3)
    Else
      edt.Text = ""
    End If
    edt.Common.Hint = "If white then enter * for automatic"
    If i < (max-1) Then
      Set edt = ui.NewButton(Sheet)
      edt.Common.SetRect 380, (i*25)+132, 50, 20
      edt.Common.ControlName = "GFButt"&i
      edt.Caption = "Clear"
      edt.UseScript = Script.ScriptPath
      edt.OnClickFunc = "MoreClear"
    End If
  WEnd
  If i < (max-1) Then
    i = i + 1
    Set edt = ui.NewDropDown(Sheet)
    edt.Common.SetRect 5, (i*25)+132, 80, 20
    edt.Common.ControlName = "GFProp"&i
    edt.Style = 2
    edt.AddItem(" ")
    edt.AddItem("Custom1")
    edt.AddItem("Custom2")
    edt.AddItem("Custom3")
    edt.AddItem("Genre")
    edt.AddItem("Mood")
    edt.AddItem("Occasion")
    edt.AddItem("Quality")
    edt.AddItem("Tempo")
    edt.AddItem("Custom4")
    edt.AddItem("Custom5")    
    edt.AddItem("Comment")
    edt.ItemIndex = 0
    Set edt = ui.NewDropDown(Sheet)
    edt.Common.SetRect 90, (i*25)+132, 60, 20
    edt.Common.ControlName = "GFType"&i
    edt.Style = 2
    edt.AddItem("None")
    edt.AddItem("White")
    edt.AddItem("Black")
    edt.ItemIndex = 0
    Set edt = ui.NewEdit(Sheet)
    edt.Common.SetRect 155, (i*25)+132, 220, 20
    edt.Common.ControlName = "GFText"&i
    edt.Text = ""
    edt.Common.Hint = "If white then enter * for automatic"
    If i < (max-1) Then
      Set edt = ui.NewButton(Sheet)
      edt.Common.SetRect 380, (i*25)+132, 50, 20
      edt.Common.ControlName = "GFButt"&i
      edt.Caption = "More"
      edt.UseScript = Script.ScriptPath
      edt.OnClickFunc = "MoreAdd"
    End If
  End If    
End Sub

Sub Event_OnSelect(drp)
  Dim par : Set par = drp.Common.TopParent.Common
  If drp.ItemIndex = 1 Then
    par.ChildControl("GFMatch").Common.Enabled = False
  Else
    par.ChildControl("GFMatch").Common.Enabled = True
  End If
End Sub

Sub MoreClear(but)
  Dim s : s = Mid(but.Common.ControlName,7)
  If IsNumeric(s) Then
    Dim i : i = Int(s)
    Dim edt : Set edt = but.Common.TopParent.Common.ChildControl("GFProp"&i)
    If Not (edt Is Nothing) Then
      edt.ItemIndex = 0
    End If
    Set edt = but.Common.TopParent.Common.ChildControl("GFType"&i)
    If Not (edt Is Nothing) Then
      edt.ItemIndex = 0
    End If    
    Set edt = but.Common.TopParent.Common.ChildControl("GFText"&i)
    If Not (edt Is Nothing) Then
      edt.Text = ""
    End If
  End If
End Sub

Sub MoreAdd(but)
  Dim s : s = Mid(but.Common.ControlName,7)
  If IsNumeric(s) Then  
    Dim i : i = Int(s)
    Dim max : max = 10
    but.Common.ControlName = "GFButt"&i
    but.Caption = "Clear"
    but.OnClickFunc = "MoreClear"    
    i = i + 1
    Dim Sheet : Set Sheet = but.Common.TopParent
    Dim ui : Set ui = SDB.UI
    Dim edt : Set edt = ui.NewDropDown(Sheet)
    edt.Common.SetRect 5, (i*25)+132, 80, 20
    edt.Common.ControlName = "GFProp"&i
    edt.Style = 2
    edt.AddItem(" ")
    edt.AddItem("Custom1")
    edt.AddItem("Custom2")
    edt.AddItem("Custom3")
    edt.AddItem("Genre")
    edt.AddItem("Mood")
    edt.AddItem("Occasion")
    edt.AddItem("Quality")
    edt.AddItem("Tempo")
    edt.AddItem("Custom4")
    edt.AddItem("Custom5")    
    edt.AddItem("Comment")
    edt.ItemIndex = 0
    Set edt = ui.NewDropDown(Sheet)
    edt.Common.SetRect 90, (i*25)+132, 60, 20
    edt.Common.ControlName = "GFType"&i
    edt.Style = 2
    edt.AddItem("None")
    edt.AddItem("White")
    edt.AddItem("Black")
    edt.ItemIndex = 0
    Set edt = ui.NewEdit(Sheet)
    edt.Common.SetRect 155, (i*25)+132, 220, 20
    edt.Common.ControlName = "GFText"&i
    edt.Text = ""
    edt.Common.Hint = "If white then enter * for automatic"
    If i < (max-1) Then
      Set edt = ui.NewButton(Sheet)
      edt.Common.SetRect 380, (i*25)+132, 50, 20
      edt.Common.ControlName = "GFButt"&i
      edt.Caption = "More"
      edt.UseScript = Script.ScriptPath
      edt.OnClickFunc = "MoreAdd"
    End If
  End If
End Sub

Sub SaveSheet(Sheet)
  Set SDB.Objects("GenreFinderCache") = CreateObject("Scripting.Dictionary")
  
  Dim ini : Set ini = SDB.IniFile
  ini.IntValue("GenreFinder","Debug") = Sheet.Common.ChildControl("GFDebug").ItemIndex
  ini.IntValue("GenreFinder","RunMode") = Sheet.Common.ChildControl("GFRunMode").ItemIndex
  ini.IntValue("GenreFinder","Confirm") = Sheet.Common.ChildControl("GFConfirm").ItemIndex
  ini.BoolValue("GenreFinder","Overwrite") = Sheet.Common.ChildControl("GFOverwrite").Checked
  ini.BoolValue("GenreFinder","OnPlay") = Sheet.Common.ChildControl("GFOnPlay").Checked    
  ini.StringValue("GenreFinder","Playlist") = Sheet.Common.ChildControl("GFPlaylist").Text
  ini.BoolValue("GenreFinder","Multiple") = Sheet.Common.ChildControl("GFMultiple").Checked 
  ini.IntValue("GenreFinder","Match") = Sheet.Common.ChildControl("GFMatch").Value
  ini.IntValue("GenreFinder","First") = Sheet.Common.ChildControl("GFFirst").Value
  ini.BoolValue("GenreFinder","Library") = Sheet.Common.ChildControl("GFLibrary").Checked 
  
  Dim but : Set but = SDB.Objects("GFToolbarButton")
  If Not (but Is Nothing) Then
    but.Visible = Sheet.Common.ChildControl("GFToolbar").Checked
  End If
  ini.BoolValue("GenreFinder","Toolbar") = Sheet.Common.ChildControl("GFToolbar").Checked
  
  Dim flds : flds = ""
  Dim i : i = 1
  Dim edt : Set edt = Sheet.Common.ChildControl("GFProp"&i)
  While Not (edt Is Nothing)
    If edt.ItemIndex > 0 Then
      Dim edt2 : Set edt2 = Sheet.Common.ChildControl("GFType"&i)
      If Not (edt2 Is Nothing) Then
        Dim edt3 : Set edt3 = Sheet.Common.ChildControl("GFText"&i)
        If Not (edt3 Is Nothing) Then
          Dim s : s = HexIt(edt.ItemIndex)&edt2.ItemIndex
          If flds = "" Then
            flds = s
          Else
            flds = flds&"|"&s
          End If
          If edt2.ItemIndex > 0 Then
            If Left(edt3.Text,1) = "*" Then
              If edt2.ItemIndex = 1 Then
                flds = flds&edt3.Text
              End If
            Else
              flds = flds&edt3.Text
            End If
          End If
        End If
      End If
    End If
    i = i + 1
    Set edt = Sheet.Common.ChildControl("GFProp"&i)
  WEnd
  ini.StringValue("GenreFinder","Lists") = flds 
End Sub

Function HexIt(str)
  HexIt = ""
  Dim dig : dig = Int(str)
  If (dig > -1) And (dig < 16) Then
    Select Case dig
      Case 10
        HexIt = "A"
      Case 11
        HexIt = "B"
      Case 12
        HexIt = "C"
      Case 13
        HexIt = "D"
      Case 14
        HexIt = "E"
      Case 15
        HexIt = "F"        
      Case Else
        HexIt = CStr(dig)
    End Select
  End If  
End Function

Function DigIt(str)
  DigIt = 0
  If Len(str) = 1 Then
    If IsNumeric(str) Then
      DigIt = Int(str)
    Else
      Select Case str
        Case "A"
          DigIt = 10
        Case "B"
          DigIt = 11
        Case "C"
          DigIt = 12
        Case "D"
          DigIt = 13          
        Case "E"
          DigIt = 14
        Case "F"
          DigIt = 15          
      End Select
    End If
  End If
End Function

Function Acceptable(temp,str)
  Acceptable = str
  If (str = "") Or (Len(temp) < 3) Then
    Exit Function
  End If

  'check list
  Dim Debug : Debug = SDB.IniFile.IntValue("GenreFinder","Debug")
  Dim fld : fld = Int(Left(temp,1))
  Dim typ : typ = Int(Mid(temp,2,1))
  Dim lst : lst = Mid(temp,3)
  Dim prp : prp = FieldToProp(fld)  
  Select Case typ
    Case 1
      Dim dic : Set dic = SDB.Objects("GenreFinderAutoLists")
      Dim res : res = ""
      If Left(lst,1) = "*" Then
        If dic.Exists(prp) Then
          lst = dic.Item(prp)
        Else
          If Len(lst) > 1 Then
            res = Mid(lst,2)
          End If
          lst = ""          
          Dim tab : tab = Left(res,1)
          If (tab = "+") Or (tab = "-") Or (tab = "") Then
            tab = CStr(fld)
          Else
            tab = GetField(res)
            res = Mid(res,Len(FieldToProp(tab))+1)
          End If     
          Dim sql : sql = FieldToSQL(tab)
          If Not (sql = "") Then
            Dim itr : Set itr = SDB.Database.OpenSQL(sql)
            While Not itr.EOF
              Dim val : val = itr.StringByIndex(0)
              If Not (val = "") Then
                val = Replace(val,"; ",",")
                If lst = "" Then
                  lst = val
                Else
                  lst = lst&","&val
                End If
              End If
              itr.Next
            WEnd
            Set itr = Nothing
          End If
          lst = ModifyList(lst,res)
          If Debug > 2 Then 
            Call out(prp&" whitelist is '"&lst&"'")
          End If
          dic.Item(prp) = lst
        End If
      Else
        If Left(lst,1) = "@" Then
          If dic.Exists(prp) Then
            lst = dic.Item(prp)
          Else
            If Len(lst) > 1 Then
              res = Mid(lst,2)
            End If
            lst = ""                        
            If Not (res = "") Then
              Dim fso : Set fso = CreateObject("Scripting.FileSystemObject")
              If fso.FileExists(res) Then
                Dim fil : Set fil = fso.OpenTextFile(res,1,False)
                lst = fil.ReadLine
                fil.Close
              End If
            End If
            If Debug > 2 Then 
              Call out(prp&" whitelist is '"&lst&"'")
            End If
            dic.Item(prp) = lst            
          End If
        End If
      End If
      Set SDB.Objects("GenreFinderAutoLists") = dic
      res = IsInList(lst,str)
      If Not (res = "[TRUE]") Then
        If res = "[FALSE]" Then
          If Debug > 1 Then 
            Call out(prp&" '"&str&"' is not in whitelist")
          End If
          Acceptable = ""
        Else
          If Debug > 1 Then 
            Call out(prp&" '"&str&"' translated to '"&res&"'.")
          End If        
          Acceptable = res
        End If
      End If    
    Case 2
      If Not (IsInList(lst,str) = "[FALSE]") Then
        If Debug > 1 Then 
          Call out(prp&" '"&str&"' is in blacklist.")
        End If        
        Acceptable = ""
      End If    
  End Select
End Function

Function IsDup(str,res)
  IsDup = True
  Dim i : i = 0
  Dim arr : arr = Split(str,"; ")
  For i = 0 To UBound(arr)
    If arr(i) = res Then
      Exit Function
    End If
  Next
  IsDup = False
End Function

Function ModifyList(orig,mods)
  ModifyList = orig
  Dim m : m = Left(mods,1)
  If (m <> "+") And (m <> "-") Then
    Exit Function
  End If
  Dim a : a = Split(orig,",")
  Dim d : Set d = CreateObject("Scripting.Dictionary")
  Dim i : i = 0
  For i = 0 To UBound(a)
    d.Item(a(i)) = i
  Next
  Dim w : w = ""
  For i = 2 To Len(mods)
    Dim c : c = Mid(mods,i,1)
    If (c = "+") Or (c = "-") Then
      If c = Mid(mods,i+1,1) Then
        w = w&c
        i = i+1
      Else
        If m = "+" Then
          d.Item(w) = m
        Else
          If d.Exists(w) Then
            d.Remove(w)
          End If
        End If
        m = c
        w = ""
      End If
    Else
      w = w&c
    End If
  Next
  If Not (w = "") Then
    If m = "+" Then
      d.Item(w) = m
    Else
      If d.Exists(w) Then
        d.Remove(w)
      End If
    End If
  End If  
  ModifyList = Join(d.Keys,",")
End Function

Function FieldToProp(ind)
  FieldToProp = ""
  Select Case DigIt(ind)
    Case 1
      FieldToProp = "Custom1"
    Case 2
      FieldToProp = "Custom2"
    Case 3
      FieldToProp = "Custom3"
    Case 4
      FieldToProp = "Genre"
    Case 5
      FieldToProp = "Mood"
    Case 6
      FieldToProp = "Occasion"
    Case 7
      FieldToProp = "Quality"
    Case 8
      FieldToProp = "Tempo"
    Case 9
      FieldToProp = "Custom4"
    Case 10
      FieldToProp = "Custom5"      
    Case 11
      FieldToProp = "Comment"      
  End Select
End Function

Function FieldToSQL(ind)
  FieldToSQL = ""
  Select Case DigIt(ind)
    Case 4
      FieldToSQL = "SELECT DISTINCT GenreName FROM Genres"
    Case Else
      FieldToSQL = FieldToProp(ind)
      If Not (FieldToSQL = "") Then
        FieldToSQL = "SELECT DISTINCT "&FieldToSQL&" FROM Songs"
      End If
  End Select
End Function

Function GetField(str)
  GetField = ""
  Dim i : i = 0
  Dim p : p = ""
  Do While (i < Len(str))
    i = i+1
    Dim c : c = Mid(str,i,1)
    If (c = "+") Or (c = "-") Or (c = "") Then
      Exit Do
    End If
    p = p&c
  Loop
  GetField = HexIt(PropToField(p))
End Function

Function PropToField(pro)
  PropToField = 0
  Select Case pro
    Case "Custom1"
      PropToField = 1
    Case "Custom2"
      PropToField = 2
    Case "Custom3"
      PropToField = 3
    Case "Genre"
      PropToField = 4
    Case "Mood"
      PropToField = 5
    Case "Occasion"
      PropToField = 6
    Case "Quality"
      PropToField = 7
    Case "Tempo"
      PropToField = 8
    Case "Custom4"
      PropToField = 9
    Case "Custom5"
      PropToField = 10      
    Case "Comment"
      PropToField = 11
  End Select
End Function

Sub Confirmation()
  Dim Debug : Debug = SDB.IniFile.IntValue("GenreFinder","Debug")
  If Debug > 0 Then
    Call out("Confirmation...")
  End If

  'get queue
  Dim que : Set que = SDB.Objects("GenreFinderQueue")
  If que Is Nothing Then
    Call ClearUp("Sorry, the queue has been lost.",mtError)
    Exit Sub
  End If  
  
  'build confirmation
  Dim Confirm : Set Confirm = SDB.Tools.FileSystem.CreateTextFile(Script.ScriptPath&".htm",True)
  Confirm.WriteLine "<html><head><title>Genre Finder</title><style type=text/css>" 
  Confirm.WriteLine "body{font-family:'Verdana',sans-serif; background-color:#FFFFFF; font-size:9pt; color:#000000;}" 
  Confirm.WriteLine "H1{font-family:'Verdana',sans-serif; font-size:10pt; font-weight:bold; color:#000000; text-align:left}"
  Confirm.WriteLine "P{font-family:'Verdana',sans-serif; font-size:8pt; color:#000000;}" 
  Confirm.WriteLine "TH{font-family:'Verdana',sans-serif; font-size:9pt; font-weight:bold; text-align:left; color:#000000; border-color:#000000; border-style: solid; border-left-width:0px; border-right-width:0px; border-top-width:0px; border-bottom-width:3px;}" 
  Confirm.WriteLine "TD{font-family:'Verdana',sans-serif; font-size:7pt; color:#000000; border-color:#000000; border-style: solid; border-left-width:0px; border-right-width:0px; border-top-width:0px; border-bottom-width:1px;}"          
  Confirm.WriteLine "</style><script language='JavaScript'>"
  Confirm.WriteLine "function toggle() {"
  Confirm.WriteLine "  var chka = document.getElementById(""chkA"")"
  Confirm.WriteLine "  var chks = document.getElementsByTagName(""input"");"
  Confirm.WriteLine "  for(var i=0;i<chks.length;i++) {"
  Confirm.WriteLine "    chks[i].checked = chka.checked;"
  Confirm.WriteLine "  }"
  Confirm.WriteLine "}"
  Select Case que.Item("run")
    Case "tracks"
      Confirm.WriteLine "</script></head><body><h1>Confirm track changes (track mode)</h1>"
    Case "artists"
      Confirm.WriteLine "</script></head><body><h1>Confirm track changes (artist mode)</h1>"
    Case "albums"
      Confirm.WriteLine "</script></head><body><h1>Confirm track changes (album mode)</h1>"
  End Select
  Confirm.WriteLine "<table width=100%><tr><th>Name</th><th>Field</th><th>Value</th><th>Change to</th><th><input type=checkbox id='chkA' onclick='toggle()' checked></th></tr>"
  Dim i : i = 0
  Dim m : m = 0
  Dim arr : arr = que.Keys
  For i = 0 To UBound(arr)
    If Left(arr(i),1) = "@" Then
      Dim id : id = Mid(arr(i),2)
      Dim dic : Set dic = que.Item(CStr(arr(i)))
      If Not (dic Is Nothing) Then
        Dim lst : Set lst = SDB.NewSongList
        Dim itm : Set itm = Nothing
        Dim nam : nam = ""
        Dim sit : Set sit = SDB.Database.QuerySongs("AND Songs.ID="&id)
        If Not (sit.EOF) Then
          Set itm = sit.Item
          nam = itm.Title&" ("&itm.ArtistName&")"   
          While Not (sit.EOF)
            Call lst.Add(sit.Item)
            sit.Next
          WEnd        
        End If
        Set sit = Nothing
        If (nam <> "") And (dic.Count > 0) Then
          Dim j : j = 0
          Dim k : k = 0
          Dim l : l = 0
          Dim txt : txt = ""          
          Dim key : key = dic.Keys
          For j = 0 To UBound(key)
            Dim tmp : Set tmp = CreateObject("Scripting.Dictionary")
            Dim val : val = dic.Item(key(j))
            Dim was : was = ""
            For k = 0 To lst.Count-1
              Execute("tmp.Item(CStr(lst.Item(k)."&key(j)&")) = k")
            Next
            Dim bla : bla = 0
            For k = 0 To tmp.Count-1 
              Dim a : a = tmp.Keys
              If a(k) = "" Then
                bla = bla + 1
              Else
                If was = "" Then
                  was = a(k)
                Else
                  was = was&", "&a(k)
                End If
              End If
            Next
            If j > 0 Then
              Confirm.WriteLine "<tr>"
            End If
            Dim boo : boo = True
            If bla = 0 Then
              If (Int(que.Item("ovr")) = 1) Or (was = val) Then
                boo = False
              End If
            End If
            If boo Then
              l = l + 1
              txt = txt&"<td>"&key(j)&"</td><td>"&MapXML(was)&"</td><td>"&MapXML(val)&"</td><td><input type=checkbox id='chk"&i&"_"&j&"' checked></td></tr>"
            End If
          Next
          If l > 0 Then
            m = m + 1
            Confirm.WriteLine "<tr><td rowspan="&l&">"&MapXML(nam)&"</td>"&txt
          End If
        End If
      End If
    End If
  Next
  Confirm.WriteLine "</table></body></html>"
  Confirm.Close
  
  'show confirmation
  If m = 0 Then
    Dim tmp2 : Set tmp2 = SDB.Objects("GenreFinderList")
    If tmp2 Is Nothing Then
      Call SDB.MessageBox("GenreFinder: All tracks have been processed - no suggested changes.",mtInformation,Array(mbOk))  
    Else
      Call SDB.MessageBox("GenreFinder: All "&tmp2.Count&" tracks have been processed - no suggested changes.",mtInformation,Array(mbOk))  
    End If
  Else
    'create the confirmation screen
    Dim Form : Set Form = SDB.UI.NewForm
    Form.Common.SetRect 50, 50, 500, 400
    Form.Common.MinWidth = 200
    Form.Common.MinHeight = 150
    Form.FormPosition = 4
    Form.Caption = "Genre Finder"
    Form.SavePositionName = "GenreFinderPosition"
    Form.StayOnTop = True
  
    Dim WB : Set WB = SDB.UI.NewActiveX(Form, "Shell.Explorer")
    WB.Common.Align = 5
    WB.Interf.Navigate Script.ScriptPath&".htm"
  
    Dim Foot : Set Foot = SDB.UI.NewPanel(Form)
    Foot.Common.Align = 2
    Foot.Common.Height = 35
  
    Dim Btn : Set Btn = SDB.UI.NewButton(Foot)
    Btn.Caption = "&"&SDB.Localize("Cancel")
    Btn.Common.Width = 85
    Btn.Common.Height = 25
    Btn.Common.Left = Form.Common.Width - Btn.Common.Width -15
    Btn.Common.Top = 6
    Btn.Common.Anchors = 2+4
    Btn.UseScript = Script.ScriptPath
    Btn.ModalResult = 2
    Btn.Cancel = True  

    Dim Btn2 : Set Btn2 = SDB.UI.NewButton(Foot)
    Btn2.Caption = "&"&SDB.Localize("Ok")
    Btn2.Common.Width = 85
    Btn2.Common.Height = 25
    Btn2.Common.Left = Btn.Common.Left - Btn2.Common.Width -5
    Btn2.Common.Top = 6
    Btn2.Common.Anchors = 2+4
    Btn2.UseScript = Script.ScriptPath
    Btn2.ModalResult = 1 
    Btn2.Default = True
  
    Dim prog : Set prog = SDB.Objects("GenreFinderProgress")
    prog.Value = 0
    prog.Text = "GenreFinder: Awaiting user confirmation..."
  
    If Form.ShowModal = 1 Then
      'save changes    
      Dim ovr : ovr = Int(que.Item("ovr"))
      prog.MaxValue = que.Count
      prog.Text = "GenreFinder: Saving selected changes..."
      For i = 0 To UBound(arr)
        prog.Value = i
        If Left(arr(i),1) = "@" Then
          Dim fld : Set fld = que.Item(CStr(arr(i)))
          If Not (fld Is Nothing) Then
            Dim iter : Set iter = SDB.Database.QuerySongs("AND Songs.ID="&Mid(arr(i),2))           
            If Not (iter.EOF) Then
              Dim text : text = iter.Item.ArtistName&" - "&iter.Item.Title
              Dim list : Set list = SDB.NewSongList
              While Not (iter.EOF)
                Dim n : n = 0
                Dim upd : upd = False
                Dim ar2 : ar2 = fld.Keys
                For n = 0 To UBound(ar2)
                  Dim ch2 : Set ch2 = WB.Interf.Document.getElementById("chk"&i&"_"&n)
                  If Not (ch2 Is Nothing) Then
                    If ch2.Checked Then
                      Dim va2 : va2 = ""
                      Execute("va2 = iter.Item."&ar2(n))
                      If (ovr = 1) And Not (va2 = "") Then
                        'don't overwrite unless blank
                      Else
                        upd = True
                        Execute("iter.Item."&ar2(n)&" = """&fld.Item(ar2(n))&"""")
                        If (Debug > 0) And (list.Count = 0) Then
                          Call out(text&": "&ar2(n)&"="&fld.Item(ar2(n)))
                        End If                      
                      End If
                    End If
                  End If
                Next
                If upd Then
                  Call list.Add(iter.Item)                 
                End If                      
                iter.Next
              WEnd
              Set iter = Nothing
              If list.Count > 0 Then
                SDB.Database.Commit
                SDB.Database.BeginTransaction
                Call list.UpdateAll()
                SDB.Database.Commit
              End If
            End If
            Set iter = Nothing   
          End If
        End If
        SDB.ProcessMessages
        If prog.Terminate Then
          If Debug > 0 Then
            Call out("(Terminated)")
          End If                
          Exit For
        End If        
      Next
      prog.Value = prog.MaxValue 
    Else
      If Debug > 0 Then
        Call out("(Cancelled)")
      End If       
    End If 
  End If
  
  'clean up
  Call SDB.Tools.FileSystem.DeleteFile(Script.ScriptPath&".htm")
  Set SDB.Objects("GenreFinderQueue") = Nothing
  Set SDB.Objects("GenreFinderProgress") = Nothing
  Set SDB.Objects("GenreFinderList") = Nothing
  Set SDB.Objects("GenreFinderItem") = Nothing
  Set SDB.Objects("GenreFinderXML") = Nothing
  Set SDB.Objects("GenreFinderPlaylist") = Nothing
  Set SDB.Objects("GenreFinderAutoLists") = Nothing
End Sub

Function MapXML(original) 
  Dim hold : hold = original 
  hold = Replace(hold,"&","&amp;") 
  hold = Replace(hold," ","&nbsp;") 
  hold = Replace(hold,"<","&lt;") 
  hold = Replace(hold,">","&gt;") 
  Dim i : i = 1 
  While i<=Len(hold) 
    If (AscW(Mid(hold,i,1))>127) Then 
      hold = Mid(hold,1,i-1)+"&#"+CStr(AscW(Mid(hold,i,1)))+";"+Mid(hold,i+1) 
    End If 
    i = i + 1 
  WEnd 
  If hold = "" Then
    MapXML = "&nbsp;"
  Else
    MapXML = hold 
  End If
End Function

Sub NextTrack(cur,nxt)
  Set nxt = Nothing
  If cur Is Nothing Then
    Exit Sub
  End If
  Dim fnd : fnd = False
  Dim sql : sql = "AND Songs.IDAlbum = "&cur.Album.ID&" ORDER BY DiscNumber COLLATE NUMERICSTRING, TrackNumber COLLATE NUMERICSTRING"
  Dim iter : Set iter = SDB.Database.QuerySongs(sql)
  Do While Not iter.EOF
    If fnd Then
      Set nxt = iter.Item
      Exit Do
    End If
    If iter.Item.ID = cur.ID Then
      fnd = True
    End If
    iter.Next
  Loop
  Set iter = Nothing
End Sub

Function FieldToAlbumSQL(ind,aid)
  FieldToAlbumSQL = ""
  Select Case DigIt(ind)
    Case 4
      FieldToAlbumSQL = "SELECT Count(Songs.ID), Genres.GenreName FROM Genres, GenresSongs, Songs WHERE Genres.IDGenre=GenresSongs.IDGenre AND GenresSongs.IDSong=Songs.ID AND Songs.IDAlbum="&aid&" GROUP BY Genres.GenreName ORDER BY Count(Songs.ID) DESC"
    Case Else
      FieldToAlbumSQL = FieldToProp(ind)
      If Not (FieldToAlbumSQL = "") Then
        FieldToAlbumSQL = "SELECT Count(Songs.ID), Songs."&FieldToAlbumSQL&" FROM Songs WHERE Songs.IDAlbum="&aid&" GROUP BY Songs."&FieldToAlbumSQL&" ORDER BY Count(Songs.ID) DESC"
      End If
  End Select
End Function

Function FieldToArtistSQL(ind,aid)
  FieldToArtistSQL = ""
  Select Case DigIt(ind)
    Case 4
      FieldToArtistSQL = "SELECT Count(Songs.ID), Genres.GenreName FROM Genres, GenresSongs, Songs, ArtistsSongs WHERE Genres.IDGenre=GenresSongs.IDGenre AND GenresSongs.IDSong=Songs.ID AND ArtistsSongs.IDSong=Songs.ID AND ArtistsSongs.IDArtist="&aid&" GROUP BY Genres.GenreName ORDER BY Count(Songs.ID) DESC"
    Case Else
      FieldToArtistSQL = FieldToProp(ind)
      If Not (FieldToArtistSQL = "") Then
        FieldToArtistSQL = "SELECT Count(Songs.ID), Songs."&FieldToArtistSQL&" FROM Songs,ArtistsSongs WHERE ArtistsSongs.IDSong=Songs.ID AND ArtistsSongs.IDArtist="&aid&" GROUP BY Songs."&FieldToArtistSQL&" ORDER BY Count(Songs.ID) DESC"
      End If
  End Select
End Function

Sub LocalSearch(flds,que,itm,aid)
  If flds = "" Then
    Exit Sub
  End If
  If que Is Nothing Then
    Exit Sub
  End If
  If itm Is Nothing Then
    Exit Sub
  End If
  
  'read library
  Dim Debug : Debug = Int(que.Item("deb"))
  Dim dic : Set dic = CreateObject("Scripting.Dictionary")
  Dim run : run = que.Item("run")
  Dim ovr : ovr = Int(que.Item("ovr"))
  Dim con : con = Int(que.Item("con"))
  Dim mul : mul = Int(que.Item("mul"))   
  Dim fir : fir = Int(que.Item("fir")) 
  If mul = 0 Then
    fir = 1
  End If
  Dim arr : arr = Split(flds,"|")
  Dim i : i = 0
  For i = 0 To UBound(arr)
    Dim j : j = 1  
    Dim fld : fld = arr(i) 
    Dim f : f = Left(fld,1)
    Dim nam : nam = FieldToProp(f)
    Dim sql : sql = ""
    Select Case run 
      Case "albums"
        sql = FieldToAlbumSQL(f,aid)
      Case "artists"
        sql = FieldToArtistSQL(f,aid)
    End Select
    Dim iter : Set iter = SDB.Database.OpenSQL(sql)
    Do While Not iter.EOF
      If (iter.ValueByIndex(0) > 1) And (iter.StringByIndex(1) <> "") Then 
        Dim str : str = FixGenre(iter.StringByIndex(1))
        Dim res : res = Acceptable(fld,str)
        If Not (res = "") Then
          If dic.Exists(nam) Then
            str = dic.Item(nam)
            If Not IsDup(str,res) Then
              dic.Item(nam) = str&"; "&res
            End If
          Else
            Call dic.Add(nam,res)
          End If
          If mul = 0 Then
            res = "|"&flds&"|"
            res = Replace(res,"|"&arr(i)&"|","|")
            If res = "|" Then
              flds = ""
            Else
              flds = Mid(res,2,Len(res)-2)
            End If
          End If
        End If    
      End If
      j = j + 1
      If (fir > 0) And (j > fir) Then
        Exit For
      End If      
      iter.Next
    Loop 
    Set iter = Nothing
  Next
  
  'cache results
  Dim cac : Set cac = SDB.Objects("GenreFinderCache")
  Select Case run
    Case "albums"
      Call cac.Add("Album"&aid,dic)
    Case "artists"
      Call cac.Add("Artist"&aid,dic)
  End Select
  
  'update track
  If flds = "" Then
    Call UpdateTrack(dic,que,itm)
  End If     
End Sub

Sub Install()
  Dim inip : inip = SDB.ApplicationPath&"Scripts\Scripts.ini"
  Dim inif : Set inif = SDB.Tools.IniFileByPath(inip)
  If Not (inif Is Nothing) Then
    inif.StringValue("GenreFinder","Filename") = "Auto\GenreFinder.vbs"
    inif.StringValue("GenreFinder","Procname") = "GenreFinder"
    inif.StringValue("GenreFinder","Order") = "50"
    inif.StringValue("GenreFinder","DisplayName") = "Genre Finder"
    inif.StringValue("GenreFinder","Description") = "Finds track information from Last.Fm"
    inif.StringValue("GenreFinder","Language") = "VBScript"
    inif.StringValue("GenreFinder","ScriptType") = "0"
    SDB.RefreshScriptItems
  End If
  Call OnStartup()
End Sub  